/* 
 * FILE --- fuzzy_ts_sys.cc
 *
 * PURPOSE --- Source file for a class for a multiple input single output 
 * (MISO) Takagi-Sugeno (TS) fuzzy system. The user specifies the number of inputs to the
 * fuzzy system, the number of the Takagi-Sugeno inputs, the number of membership 
 * functions (MF) on each universe of discourse (UD). The MF can be defined as triangular 
 * with 50% of overlapp or as gaussian. Either minimum or product can be used for 
 * premises whereas the defuzzification method is fixed as singleton defuzzification.
 * The utility has a build-in adaptation capability of the output parameters (TS coefficients)
 * based on the Normalized Gradient Method or the so called Projection Algorithm.
 * Therefore it can be used in both direct and indirect adaptive control schemes as
 * a function approximator.
 *
 * This utility uses the matrix tool for matrix maltiplications.
 *
 * Uses two other utilities:
 *                  FUZZY_UD --- Fuzzy universe of discourse
 *                  FUZZY_MF --- Fuzzy membership function
 *                  MATRIX --- Fuzzy membership function
 *
 */

// Include Files
#include <math.h>
#include "fuzzy_ts_sys.hh"
#include "random.hh"

// Define this macro if you want to use this utility without linking
// to the RCS Library
#ifdef NO_RCS
#include <stdio.h>            /* printf function */
#else
#include "rcs_prnt.hh"          /* rcs_print_ functions */
#endif


/*************************************************************/

FUZZY_TS_SYS::FUZZY_TS_SYS(int _num_of_inputs, 
			   int _num_of_ts_inputs,
			   int *_num_of_in_mfs = NULL)
{
  // No error at start
  fuzzy_ts_sys_err = FUZZY_TS_SYS_NO_ERROR;
  num_of_inputs = _num_of_inputs;
  num_of_ts_inputs = _num_of_ts_inputs;
  
  // Allocate memory 
  allocate_memory(_num_of_in_mfs);
  
  // If error then report
  if(fuzzy_ts_sys_err != FUZZY_TS_SYS_NO_ERROR)
    {
#ifdef NO_RCS
      printf("FUZZY_TS_SYS: ERROR --- Not Enought Memory!!!\n");
      printf("The Fuzzy System will not function correctly ...\n");
#else
      rcs_print("FUZZY_TS_SYS: ERROR --- Not Enought Memory!!!\n");
      rcs_print("The Fuzzy System will not function correctly ...\n");
#endif
    }
  
  // Set the the default premise type
  set_premise_type();
  
  // Set the input gains to zeros
  set_input_gains();

  // Initialize the adaptation gain to 1.0
  eta = 1.0;

  // Phi norm is zero initially
  phi_norm = 0.0;

  // No infrence calculated yet
  total_inference = 0.0;
}


// Destructor
FUZZY_TS_SYS::~FUZZY_TS_SYS()
{
  // Free the allocated memory
  free_memory();
}

void FUZZY_TS_SYS::allocate_memory(int *_num_of_in_mfs)
{
  int ii;  // Index variales for the loops
  
  // Input MF 
  if ((in_ud = new fuzzy_ud*[num_of_inputs]) == NULL)
    fuzzy_ts_sys_err = FUZZY_TS_SYS_MEM_ERROR;

  if (_num_of_in_mfs == NULL) // No num of MF is specified
    {
      for (ii=0; ii < num_of_inputs; ii++)
	{
	  in_ud[ii] = new fuzzy_ud(DEFAULT_NUM_IN_MF);
	  if (in_ud[ii]->fuzzy_ud_err != FUZZY_UD_NO_ERROR)
	    fuzzy_ts_sys_err = FUZZY_TS_SYS_MEM_ERROR;
	}
    }
  else  // Num of MF is specified
    {
      for (ii=0; ii < num_of_inputs; ii++)
	{
	  in_ud[ii] =  new fuzzy_ud(_num_of_in_mfs[ii]);
	  if (in_ud[ii]->fuzzy_ud_err != FUZZY_UD_NO_ERROR)
	    fuzzy_ts_sys_err = FUZZY_TS_SYS_MEM_ERROR;
	}
    }
  
  // Calculate the # of rules 
  num_of_rules = 1;
  for (ii = 0; ii < num_of_inputs; ii++)
    {
      num_of_rules *= in_ud[ii]->get_num_of_mfs();
    }

  // Calculate the # of output parameters
  num_of_out_params = num_of_rules*(num_of_ts_inputs + 1);
  
  // Takagi-Sugeno coefficients
  if ((ts_coeff = new matrix(num_of_rules, num_of_ts_inputs+1)) == NULL)
    fuzzy_ts_sys_err = FUZZY_TS_SYS_MEM_ERROR;
  
  // Conveniance variable
  if ((ts_inputs = new matrix(num_of_ts_inputs+1, 1)) == NULL)
    fuzzy_ts_sys_err = FUZZY_TS_SYS_MEM_ERROR;

  // Conveniance variable
  if ((inference = new matrix(1, num_of_rules)) == NULL)
    fuzzy_ts_sys_err = FUZZY_TS_SYS_MEM_ERROR;


  // Create the vector of output parametrs
  theta = new matrix(1, num_of_out_params);
  if (theta->matrix_err != MATRIX_NO_ERROR)
    {
      fuzzy_ts_sys_err = FUZZY_TS_SYS_MEM_ERROR;
    }

  // Create the regression vector
  phi = new matrix(num_of_out_params, 1);
  if (phi->matrix_err != MATRIX_NO_ERROR)
    {
      fuzzy_ts_sys_err = FUZZY_TS_SYS_MEM_ERROR;
    }
}

void FUZZY_TS_SYS::free_memory(void)
{
  int ii;  // Index variales for the loops
  
  for (ii = 0; ii < num_of_inputs; ii++)
    {
      delete in_ud[ii];
    }
  delete [] in_ud;
  
  delete ts_coeff;
  delete ts_inputs;
  delete inference;

  delete theta;
  delete phi;
}


// This function sets the premise type. Invalid premices are considered
// as MINIMUM_PREMISE.
void FUZZY_TS_SYS::set_premise_type(int _premise_type = MINIMUM_PREMISE)
{
  premise_type = _premise_type;
  if((premise_type != MINIMUM_PREMISE) && (premise_type != PRODUCT_PREMISE))
    {
#ifdef NO_RCS
      printf("Invalid premise type!!!\n");
      printf("Minimum premise assumed ...\n");
#else
      rcs_print("Invalid premise type!!!\n");
      rcs_print("Minimum premise assumed ...\n");
#endif
      premise_type = MINIMUM_PREMISE;      
    }
}

void FUZZY_TS_SYS::set_input_gains(double *_in_gains = NULL)
{
  int ii;  // Index variales for the loops
  
  // Set the gains for each input universe
  if (_in_gains == NULL) // No gains specified
    {
      for (ii = 0; ii < num_of_inputs; ii++)
	{
	  in_ud[ii]->ud_gain = 1.0;
	}
    }
  else // Gains specified
    {
      for (ii=0; ii < num_of_inputs; ii++)
	{
	  in_ud[ii]->ud_gain = _in_gains[ii];
	}
    }
}

/* Set the Takagi-Sugeno coefficients to the specified values */
void FUZZY_TS_SYS::set_ts_coeff(const double **_ts_coeff)
{
  // Set the TS coefficients matrix to the specified value
  ts_coeff->set_matrix(_ts_coeff);
}

/* Set the Takagi-Sugeno coefficients to the specified values */
void FUZZY_TS_SYS::set_ts_coeff(const double *_ts_coeff)
{
  // Set the TS coefficients matrix to the specified value
  ts_coeff->set_matrix(_ts_coeff);
}

void FUZZY_TS_SYS::set_ts_coeff(const double _ts_coeff)
{
  // Set the TS coefficients matrix to the specified value
  ts_coeff->set_matrix(_ts_coeff);
}


void FUZZY_TS_SYS::set_ts_coeff(double _low, double _high)
{
  int ii, jj;  // Index variales for the loops
  double rand_val;  // Temporary variable

  // Set the output parametrs random values b/n [_low, _high]
  for (ii = 0; ii < num_of_rules; ii++)
    {
      for (jj = 0; jj < num_of_ts_inputs+1; jj++)
        {
           // Generate a random number b/n [_low, _high]
           rand_val = Rand_Double(_low, _high);
           
           // Assign the element to the generated value
           ts_coeff->set_element(ii, jj, rand_val);
         }
    }
}


/*
  The function that given an array of crisp inputs evaluates the crisp 
  output of the fuzzy controller. This is the only function to be called 
  from outside.
  input_ = an array of crisp inputs to the premices of the fuzzy system
  _ts_input = an array TS inputs to the rule outputs of the TS fuzzy system
*/
double FUZZY_TS_SYS::control(const double *_input, const double *_ts_input)
{
  // Fuzzyfy the input
  fuzzyfy(_input);
  
  // Assign the Takagi-Sugeno inputs vector to use the matrix operations
  set_ts_input_vector(_ts_input);

  // Evaluate and return the crisp outputs
  return(defuzzyfy());
}  

  
double FUZZY_TS_SYS::defuzzyfy()
{
  int ii, jj;  // Index variales for the loops
  int input_flag, on_off_index;
  double rule_inference;
  
  // Cumulative variable: Start with zero
  total_inference = 0.0;

  /*
    The algorithm below evaluates all the possible rules.
    With the help of input_flag and on_off_index we find
    which fuzzy set (i.e., MF) from which input belongs
    to each rule.
  */
  for (ii = 0; ii < num_of_rules; ii++)
    {
      input_flag = ii; // In each rule belongs different MF from each UD
      rule_inference = 1.0; // Maximum possible inference
      
      for (jj = 0; jj < num_of_inputs; jj++) 
	{
	  on_off_index = (input_flag % in_ud[jj]->get_num_of_mfs());
	  input_flag = 
	    (input_flag - on_off_index)/in_ud[jj]->get_num_of_mfs();
	  
	  // Find the inference
	  switch(premise_type)
	    {
	    case MINIMUM_PREMISE:
	      rule_inference = 
		MIN(in_ud[jj]->deg_of_mbrship[on_off_index], rule_inference);
	      break;
	    case PRODUCT_PREMISE:
	      rule_inference *= in_ud[jj]->deg_of_mbrship[on_off_index];
	      break;
	    default:
	      // This wont happen, but just in case ...
	      // Play Safe --- Consider every other type as MINIMUM_PREMISE
	      rule_inference = 
		MIN(in_ud[jj]->deg_of_mbrship[on_off_index], rule_inference);
	    }
	}
      
      // Calculate the inference of every rule
      inference->set_element(0, ii, rule_inference);
      total_inference += rule_inference;
    }

  if (total_inference <= 0)
    {
#ifdef NO_RCS
      printf("Inproper fuzzy system!!!\n");
      printf("Sum of premise certainties equal to zero!!!\n");
      printf("Setting output %d to zero ...\n", ii);
#else
      rcs_print("Inproper fuzzy system!!!\n");
      rcs_print("Sum of premise certainties equal to zero!!!\n");
      rcs_print("Setting output %d to zero ...\n", ii);
#endif
      return(0.0);
    }
  else
    {
      return((1.0/total_inference)*
	     ((*inference)*(*ts_coeff)*(*ts_inputs)).get_element(0,0));
    }
}

void FUZZY_TS_SYS::update_ts_coeff(const double error)
{
  // Copy the TS coefficients from matrix A to vector theta
  // for the update algorithm
  A2theta();

  // Calculate the regression vector for the update algorithm
  get_phi();

  // Update the parameters based on the Normilized Gradient Method
  // or the so called Projection Algorithm
  (*theta) = (*theta) + (eta/(1.0 + phi_norm)) * error * transpose((*phi));

  // Copy the updated TS coefficients from the vector theta back to matrix A
  theta2A();
}

void FUZZY_TS_SYS::update_ts_coeff(const double error, double norm2)
{
  // Copy the TS coefficients from matrix A to vector theta
  // for the update algorithm
  A2theta();

  // Calculate the regression vector for the update algorithm
  get_phi();

  // Update the parameters based on the Normilized Gradient Method
  // or the so called Projection Algorithm
  (*theta) = (*theta) + (eta/(1.0 + norm2 + phi_norm)) * error * transpose((*phi));

  // Copy the updated TS coefficients from the vector theta back to matrix A
  theta2A();
}


void FUZZY_TS_SYS::update_ts_coeff(double error, double norm2, double uu)
{
  // Copy the TS coefficients from matrix A to vector theta
  // for the update algorithm
  A2theta();

  // Calculate the regression vector for the update algorithm
  get_phi();

  // Update the parameters based on the Normilized Gradient Method
  // or the so called Projection Algorithm
  (*theta) = (*theta) + 
    ((eta*uu)/(1.0 + norm2 + uu*uu*phi_norm)) * error * transpose((*phi));

  // Copy the updated TS coefficients from the vector theta back to matrix A
  theta2A();
}


void FUZZY_TS_SYS::bound_ts_coeff(double _low, double _high)
{
  int ii, jj;  // Index variales for the loops

  // Saturates the TS coefficient values b/n [_low, _high]
  for (ii = 0; ii < num_of_ts_inputs+1; ii++)
    {
      for (jj = 0; jj < num_of_rules; jj++)
	{
          if (ts_coeff->get_element(jj, ii) < _low)
            {
	      ts_coeff->set_element(jj, ii, _low);
	    }
          else if (ts_coeff->get_element(jj, ii) > _high)
            {
	      ts_coeff->set_element(jj, ii, _high);
	    }
	}
    }
}

inline void FUZZY_TS_SYS::get_phi(void)
{
  int ii, jj;  // Index variales for the loops

  // If improper then set the regression vector to zero
  if (total_inference <= 0)
    {
      for (ii = 0; ii < num_of_out_params; ii++)
	{
	  phi->set_element(ii, 0, 0.0);

	  phi_norm = 0.0;
	}
    }
  else
    {
      for (ii = 0; ii < num_of_ts_inputs+1; ii++)
	{
	  for (jj = 0; jj < num_of_rules; jj++)
	    {
	      phi->set_element(ii*num_of_rules + jj, 0, (1.0/total_inference)*
			       inference->get_element(0, jj)*ts_inputs->get_element(ii, 0));

	      // Calculate the norm of phi (needed in indirect adaptive control)
	      phi_norm = (transpose((*phi)) * (*phi)).get_element(0, 0);
	    }
	}
    }
}


inline void FUZZY_TS_SYS::A2theta(void)
{
  int ii, jj;  // Index variales for the loops

  // Transfer the parameters from matrix A to vector theta 
  // (sutable for adaptation)
  for (ii = 0; ii < num_of_ts_inputs+1; ii++)
    {
      for (jj = 0; jj < num_of_rules; jj++)
	{
	  theta->set_element(0, ii*num_of_rules + jj, ts_coeff->get_element(jj, ii));
	}
    }
}


inline void FUZZY_TS_SYS::theta2A(void)
{
  int ii, jj;  // Index variales for the loops

  // Transfer the parameters from vector theta back to matrix A
  // (sutable for calculating the system output)
  for (ii = 0; ii < num_of_ts_inputs+1; ii++)
    {
      for (jj = 0; jj < num_of_rules; jj++)
	{
	  ts_coeff->set_element(jj, ii, theta->get_element(0, ii*num_of_rules + jj));
	}
    }
}


inline void FUZZY_TS_SYS::fuzzyfy(const double *_input)
{
  // Scale and saturate the input to the correct universe
  for (int ii = 0; ii < num_of_inputs; ii++)
    {
      in_ud[ii]->fuzzyfy(_input[ii]);
    }
}

inline void FUZZY_TS_SYS::set_ts_input_vector(const double *_ts_input)
{
  // Augment the first element with one to incorporate a constant
  ts_inputs->set_element(0, 0, 1.0); 

  // Set the rest to the given inputs
  for (int ii = 1; ii <= num_of_ts_inputs; ii++)
    {
      ts_inputs->set_element(ii, 0, _ts_input[ii-1]);
    }
}



